import socket
import asyncio
import logging
import time

logger = logging.getLogger('elbus.client')

GREETINGS = 0xEB

PROTOCOL_VERSION = 1

OP_NOP = 0
OP_PUBLISH = 1
OP_SUBSCRIBE = 2
OP_UNSUBSCRIBE = 3
OP_MESSAGE = 0x12
OP_BROADCAST = 0x13
OP_ACK = 0xFE

RESPONSE_OK = 0x01
ERR_CLIENT_NOT_REGISTERED = 0x71
ERR_DATA = 0x72
ERR_IO = 0x73
ERR_OTHER = 0x74
ERR_NOT_SUPPORTED = 0x75
ERR_BUSY = 0x76
ERR_NOT_DELIVERED = 0x77
ERR_TIMEOUT = 0x78

PING_FRAME = b'\x00' * 9


async def on_frame_default(frame):
    pass


class Client:

    def __init__(self, path, name):
        self.path = path
        self.writer = None
        self.reader_fut = None
        self.pinger_fut = None
        self.buf_size = 8192
        self.name = name
        self.frame_id = 0
        self.ping_interval = 1
        self.on_frame = on_frame_default
        self.socket_lock = asyncio.Lock()
        self.mgmt_lock = asyncio.Lock()
        self.connected = False
        self.frames = {}
        self.timeout = 1

    async def connect(self):
        with await self.mgmt_lock:
            if self.path.endswith('.sock') or self.path.endswith(
                    '.socket') or self.path.startswith('/'):
                reader, writer = await asyncio.open_unix_connection(
                    self.path, limit=self.buf_size)
            else:
                host, port = self.path.rsplit(':', maxsplit=2)
                reader, writer = await asyncio.open_connection(
                    host, int(port), limit=self.buf_size)
            buf = await asyncio.wait_for(self.readexactly(reader, 3),
                                         timeout=self.timeout)
            if buf[0] != GREETINGS:
                raise RuntimeError('Unsupported protocol')
            if int.from_bytes(buf[1:3], 'little') != PROTOCOL_VERSION:
                raise RuntimeError('Unsupported protocol version')
            writer.write(buf)
            await asyncio.wait_for(writer.drain(), timeout=self.timeout)
            buf = await asyncio.wait_for(self.readexactly(reader, 1),
                                         timeout=self.timeout)
            if buf[0] != RESPONSE_OK:
                raise RuntimeError(f'Server response: {hex(buf[0])}')
            name = self.name.encode()
            writer.write(len(name).to_bytes(2, 'little') + name)
            await asyncio.wait_for(writer.drain(), timeout=self.timeout)
            buf = await asyncio.wait_for(self.readexactly(reader, 1),
                                         timeout=self.timeout)
            if buf[0] != RESPONSE_OK:
                raise RuntimeError(f'Server response: {hex(buf[0])}')
            self.writer = writer
            self.connected = True
            self.pinger_fut = asyncio.ensure_future(self._t_pinger())
            self.reader_fut = asyncio.ensure_future(self._t_reader(reader))

    async def handle_daemon_exception(self, e):
        with await self.mgmt_lock:
            if self.connected:
                await self._disconnect()
                raise e

    async def _t_pinger(self):
        try:
            while True:
                await asyncio.sleep(self.ping_interval)
                with await self.socket_lock:
                    self.writer.write(PING_FRAME)
                    await asyncio.wait_for(self.writer.drain(),
                                           timeout=self.timeout)
        except Exception as e:
            asyncio.ensure_future(self.handle_daemon_exception(e))

    async def _t_reader(self, reader):
        try:
            while True:
                buf = await self.readexactly(reader, 6)
                if buf[0] == OP_NOP:
                    continue
                elif buf[0] == OP_ACK:
                    op_id = int.from_bytes(buf[1:5], 'little')
                    try:
                        o = self.frames.pop(op_id)
                        o.result = buf[5]
                        o.completed.set()
                    except KeyError:
                        logger.warning(f'orphaned elbus frame ack {op_id}')
                else:

                    async def read_frame(tp, data_len):
                        frame = Frame()
                        frame.type = tp
                        sender = await reader.readuntil(b'\x00')
                        data_len -= len(sender)
                        frame.sender = sender[:-1].decode()
                        if buf[0] == OP_PUBLISH:
                            topic = await reader.readuntil(b'\x00')
                            data_len -= len(topic)
                            frame.topic = topic[:-1].decode()
                        else:
                            frame.topic = None
                        data = b''
                        while len(data) < data_len:
                            buf_size = data_len - len(data)
                            prev_len = len(data)
                            try:
                                data += await reader.readexactly(
                                    buf_size if buf_size < self.buf_size else
                                    self.buf_size)
                            except asyncio.IncompleteReadError:
                                pass
                        frame.payload = data
                        return frame

                    try:
                        data_len = int.from_bytes(buf[1:5], 'little')
                        frame = await read_frame(buf[0], data_len)
                    except Exception as e:
                        logger.error(f'Invalid frame from the server: {e}')
                        raise
                    asyncio.ensure_future(self.on_frame(frame))
        except Exception as e:
            asyncio.ensure_future(self.handle_daemon_exception(e))

    async def readexactly(self, reader, data_len):
        data = b''
        while len(data) < data_len:
            buf_size = data_len - len(data)
            try:
                chunk = await reader.readexactly(
                    buf_size if buf_size < self.buf_size else self.buf_size)
                data += chunk
            except asyncio.IncompleteReadError:
                await asyncio.sleep(0.01)
        return data

    async def disconnect(self):
        with await self.mgmt_lock:
            await self._disconnect()

    async def _disconnect(self):
        self.connected = False
        self.writer.close()
        if self.reader_fut is not None:
            self.reader_fut.cancel()
        if self.pinger_fut is not None:
            self.pinger_fut.cancel()

    async def send(self, target=None, frame=None):
        try:
            with await self.socket_lock:
                self.frame_id += 1
                if self.frame_id > 0xffff_ffff:
                    self.frame_id = 1
                frame_id = self.frame_id
                o = ClientFrame(frame.qos)
                if frame.qos > 0:
                    self.frames[frame_id] = o
                flags = frame.type | frame.qos << 6
                if frame.type == OP_SUBSCRIBE or frame.type == OP_UNSUBSCRIBE:
                    topics = frame.topic if isinstance(frame.topic,
                                                       list) else [frame.topic]
                    payload = b'\x00'.join(t.encode() for t in topics)
                    self.writer.write(
                        frame_id.to_bytes(4, 'little') +
                        flags.to_bytes(1, 'little') +
                        len(payload).to_bytes(4, 'little') + payload)
                else:
                    frame_len = len(target) + len(frame.payload) + 1
                    if frame.header is not None:
                        frame_len += len(frame.header)
                    if frame_len > 0xffff_ffff:
                        raise ValueError('frame too large')
                    self.writer.write(
                        frame_id.to_bytes(4, 'little') +
                        flags.to_bytes(1, 'little') +
                        frame_len.to_bytes(4, 'little') + target.encode() +
                        b'\x00' +
                        (frame.header if frame.header is not None else b''))
                    self.writer.write(frame.payload.encode(
                    ) if isinstance(frame.payload, str) else frame.payload)
                await self.writer.drain()
                return o
        except:
            try:
                del self.frames[frame_id]
            except KeyError:
                pass
            raise

    def subscribe(self, topics):
        frame = Frame(tp=OP_SUBSCRIBE)
        frame.topic = topics
        return self.send(None, frame)

    def unsubscribe(self, topics):
        frame = Frame(tp=OP_UNSUBSCRIBE)
        frame.topic = topics
        return self.send(None, frame)

    def is_connected(self):
        return self.connected


class ClientFrame:

    def __init__(self, qos):
        self.qos = qos
        self.result = 0
        if qos > 0:
            self.completed = asyncio.Event()
        else:
            self.completed = None

    def is_completed(self):
        if self.qos > 0:
            return self.completed.is_set()
        else:
            return True

    async def wait_completed(self, timeout=None):
        if self.qos == 0:
            return RESPONSE_OK
        elif timeout:
            await asyncio.wait_for(self.completed.wait(), timeout=timeout)
        else:
            await self.completed.wait()
        return self.result


class Frame:

    def __init__(self, payload=None, tp=OP_MESSAGE, qos=1):
        self.payload = payload
        # used for zero-copy
        self.header = None
        self.type = tp
        self.qos = qos



import asyncio
import logging

from .client import OP_MESSAGE
from .client import ERR_IO, ERR_TIMEOUT, RESPONSE_OK
from .client import on_frame_default

RPC_NOTIFICATION_HEADER = b'\x00'
RPC_REQUEST_HEADER = b'\x01'
RPC_REPLY_HEADER = b'\x11'
RPC_ERROR_REPLY_HEADER = b'\x12'

RPC_NOTIFICATION = 0x00
RPC_REQUEST = 0x01
RPC_REPLY = 0x11
RPC_ERROR = 0x12

RPC_ERROR_CODE_PARSE = -32700
RPC_ERROR_CODE_INVALID_REQUEST = -32600
RPC_ERROR_CODE_METHOD_NOT_FOUND = -32601
RPC_ERROR_CODE_INVALID_METHOD_PARAMS = -32602
RPC_ERROR_CODE_INTERNAL = -32603

logger = logging.getLogger('elbus.rpc')


async def on_call_default(event):
    raise RpcException('RPC Engine not initialized',
                       RPC_ERROR_CODE_METHOD_NOT_FOUND)


async def on_notification_default(event):
    pass


class RpcException(Exception):

    def __init__(self, msg='', code=RPC_ERROR_CODE_INTERNAL):
        self.rpc_error_code = code
        self.rpc_error_payload = msg
        super().__init__(msg if isinstance(msg, str) else msg.decode())

    def __str__(self):
        return super().__str__() + f' (code: {self.rpc_error_code})'


class Rpc:

    def __init__(self, client):
        self.client = client
        self.client.on_frame = self._handle_frame
        self.call_id = 0
        self.call_lock = asyncio.Lock()
        self.calls = {}
        self.on_frame = on_frame_default
        self.on_call = on_call_default
        self.on_notification = on_notification_default

    def is_connected(self):
        return self.client.connected

    def notify(self, target, notification):
        return self.client.send(target, notification)

    def call0(self, target, request):
        request.header = RPC_REQUEST_HEADER + b'\x00\x00\x00\x00' + \
                request.method + b'\x00'
        return self.client.send(target, request)

    async def call(self, target, request):
        with await self.call_lock:
            call_id = self.call_id + 1
            if call_id == 0xffff_ffff:
                self.call_id = 0
            else:
                self.call_id = call_id
        call_event = RpcCallEvent()
        self.calls[call_id] = call_event
        request.header = RPC_REQUEST_HEADER + call_id.to_bytes(
            4, 'little') + request.method + b'\x00'
        try:
            try:
                code = await (await self.client.send(
                    target,
                    request)).wait_completed(timeout=self.client.timeout)
                if code != RESPONSE_OK:
                    try:
                        del self.calls[call_id]
                    except KeyError:
                        pass
                    err_code = -32000 - code
                    call_event.error = RpcException(
                        f'RPC error code {err_code}', code=err_code)
                    call_event.completed.set()
            except asyncio.TimeoutError:
                try:
                    del self.calls[call_id]
                except KeyError:
                    pass
                err_code = -32000 - ERR_TIMEOUT
                call_event.error = RpcException(f'RPC timeout', code=err_code)
                call_event.completed.set()
        except Exception as e:
            try:
                del self.calls[call_id]
            except KeyError:
                pass
            call_event.error = RpcException(str(e), code=-32000 - ERR_IO)
            call_event.completed.set()
        return call_event

    async def _handle_frame(self, frame):
        try:
            if frame.type == OP_MESSAGE:
                if frame.payload[0] == RPC_NOTIFICATION:
                    event = Event(RPC_NOTIFICATION, frame, 1)
                    await self.on_notification(event)
                elif frame.payload[0] == RPC_REQUEST:
                    sender = frame.sender
                    call_id_b = frame.payload[1:5]
                    call_id = int.from_bytes(call_id_b, 'little')
                    method = frame.payload[5:5 +
                                           frame.payload[5:].index(b'\x00')]
                    event = Event(RPC_REQUEST, frame, 6 + len(method))
                    event.call_id = call_id
                    event.method = method
                    if call_id == 0:
                        await self.on_call(event)
                    else:
                        reply = Reply()
                        try:
                            reply.payload = await self.on_call(event)
                            if reply.payload is None:
                                reply.payload = b''
                            reply.header = RPC_REPLY_HEADER + call_id_b
                        except Exception as e:
                            try:
                                code = e.rpc_error_code
                            except AttributeError:
                                code = RPC_ERROR_CODE_INTERNAL
                            reply.header = (
                                RPC_ERROR_REPLY_HEADER + call_id_b +
                                code.to_bytes(2, 'little', signed=True))
                            reply.payload = str(e)
                        await self.client.send(sender, reply)
                elif frame.payload[0] == RPC_REPLY or frame.payload[
                        0] == RPC_ERROR:
                    call_id = int.from_bytes(frame.payload[1:5], 'little')
                    try:
                        call_event = self.calls.pop(call_id)
                        call_event.frame = frame
                        if frame.payload[0] == RPC_ERROR:
                            err_code = int.from_bytes(frame.payload[5:7],
                                                      'little',
                                                      signed=True)
                            call_event.error = RpcException(frame.payload[7:],
                                                            code=err_code)
                        call_event.completed.set()
                    except KeyError:
                        logger.warning(f'orphaned RPC response: {call_id}')
                else:
                    logger.error(f'Invalid RPC frame code: {frame.payload[0]}')
            else:
                await self.on_frame(frame)
        except Exception as e:
            import traceback
            logger.error(traceback.format_exc())


class RpcCallEvent:

    def __init__(self):
        self.frame = None
        self.error = None
        self.completed = asyncio.Event()

    def is_completed(self):
        return self.completed.is_set()

    async def wait_completed(self, timeout=None):
        if timeout:
            if not await asyncio.wait_for(self.completed.wait(),
                                          timeout=timeout):
                raise TimeoutError
        else:
            await self.completed.wait()
        if self.error is None:
            return self
        else:
            raise self.error

    def get_payload(self):
        return self.frame.payload[5:]


class Event:

    def __init__(self, tp, frame, payload_pos):
        self.tp = tp
        self.frame = frame
        self._payload_pos = payload_pos

    def get_payload(self):
        return self.frame.payload[self._payload_pos:]


class Notification:

    def __init__(self, payload=b''):
        self.payload = payload
        self.type = OP_MESSAGE
        self.qos = 1
        self.header = RPC_NOTIFICATION_HEADER


class Request:

    def __init__(self, method, params=b''):
        self.payload = params
        self.type = OP_MESSAGE
        self.qos = 1
        self.method = method.encode() if isinstance(method, str) else method
        self.header = None


class Reply:

    def __init__(self, result=b''):
        self.payload = result
        self.type = OP_MESSAGE
        self.qos = 1
        self.header = None
